const isString = require('lodash.isstring');
const stylelint = require('stylelint');
const { report, ruleMessages, validateOptions } = stylelint.utils;

// Plugin name
const ruleName = 'sds-stylelint-plugin/namespace-pattern';

// Custom message for report
const messages = ruleMessages(ruleName, {
  expected: (prop, ns) =>
    `Expected "${prop}" to begin with '--${ns}-g-', '--${ns}-s-', or '--${ns}-c-' and be kebab-case"`,
});

module.exports = stylelint.createPlugin(ruleName, function getPlugin(actual, options) {
  return function lint(postcssRoot, postcssResult) {
    const validOptions = validateOptions(
      postcssResult,
      ruleName,
      { actual },
      {
        actual: options,
        possible: {
          namespace: [isString],
        },
        optional: true,
      },
    );

    if (!validOptions) {
      // If the options are invalid, don't lint
      return;
    }

    // Set SDS as default
    const namespaceNormalized = isString(options.namespace) ? options.namespace.toLowerCase() : 'sds';

    // Needed to parse out custom properties when used as fallbacks
    const customPropPattern = new RegExp('(--([a-z]*)-([a-z][a-z0-9]*)(-[a-z0-9]+)*)', 'g');
    // Regex to validate --{namespace}-[g|s|c]-* and kebab-case
    const sdsPattern = new RegExp(`(--${namespaceNormalized}-[g|s|c]-([a-z][a-z0-9]*)(-[a-z0-9]+)*)`);

    postcssRoot.walkDecls((decl) => {
      const prop = decl.prop;
      const value = decl.value;

      // This conditional deals with custom properties authored as a property
      // First check that the CSS property is a custom property
      if (prop.startsWith('--')) {
        // If property matches sdsPattern then don't lint
        if (sdsPattern.test(prop)) {
          return;
        }

        // Report custom properties that don't match valid pattern
        report({
          ruleName,
          result: postcssResult,
          message: messages.expected(prop, namespaceNormalized), // Build the reported message
          node: decl, // Specify the reported node
        });
      }

      // This conditional deals with custom properties authored as a value
      // First check that the value declaration starts with a var() function
      if (value.startsWith('var(')) {
        // Parse out CSS properties declared as a fallback value and store in array
        const allPropsAsValues = value.match(customPropPattern);
        // Then filter property out of array as a single value to run regex against
        const prop = allPropsAsValues.find((value) => customPropPattern.test(value));

        // If property matches sdsPattern then don't lint
        if (sdsPattern.test(prop)) {
          return;
        }

        // Report custom properties that don't match valid pattern
        report({
          ruleName,
          result: postcssResult,
          message: messages.expected(prop, namespaceNormalized), // Build the reported message
          node: decl, // Specify the reported node
        });
      }
    });
  };
});

module.exports.ruleName = ruleName;
module.exports.messages = messages;
